from __future__ import annotations

import base64
import struct
from datetime import datetime
from datetime import timezone
from typing import Any

import pylogram

from ks_session_manager.managers.base_manager import BaseSessionFileManager
from ks_session_manager.types import SessionData


class PylogramSessionFileManager(BaseSessionFileManager):
    @property
    def session_file_extension(self) -> str:
        return pylogram.storage.FileStorage.FILE_EXTENSION

    def _get_file_storage(self, *, exists: bool = False) -> pylogram.storage.SQLiteStorage:
        if not exists and self.session_path.exists():
            raise FileExistsError(f"Session storage is already exists in {self.session_path}")

        if exists and not self.session_path.exists():
            raise FileNotFoundError(f"Session storage not found in {self.session_path}")

        if self.session_path.is_dir():
            raise IsADirectoryError(f"Session file path {self.session_path} points to a directory")

        return pylogram.storage.FileStorage(name=self.session_name, workdir=self.workdir)

    async def read_session_storage(self) -> SessionData:
        storage = self._get_file_storage(exists=True)

        await storage.open()
        dc_id = await storage.dc_id()
        user_id = await storage.user_id()
        auth_key = await storage.auth_key()
        is_bot = await storage.is_bot()
        test_mode = await storage.test_mode()
        server_address = None
        server_port = None
        takeout_id = None
        await storage.save()
        await storage.close()

        return SessionData(
            dc_id=dc_id,
            auth_key=auth_key,
            user_id=user_id,
            is_bot=is_bot,
            test_mode=test_mode,
            server_address=server_address,
            server_port=server_port,
            takeout_id=takeout_id,
        )

    async def write_session_storage(
            self,
            data: SessionData,
            *,
            in_memory: bool = False,
            exist_ok: bool = False
    ) -> Any:
        if not bool(data.user_id):
            raise ValueError("user_id must be set in session data")

        if in_memory:
            storage = pylogram.storage.MemoryStorage(
                self.session_name,
                session_string=self.pack_session_data_to_string(data)
            )
        else:
            try:
                storage = self._get_file_storage(exists=False)
            except FileExistsError:
                if exist_ok:
                    return self._get_file_storage(exists=True)

                raise
            else:
                await storage.open()
                await storage.dc_id(data.dc_id)
                await storage.api_id(0)
                await storage.auth_key(data.auth_key)
                await storage.user_id(data.user_id)
                await storage.test_mode(data.test_mode)
                await storage.is_bot(data.is_bot)
                await storage.date(int(datetime.now(timezone.utc).timestamp()))
                await storage.save()
                await storage.close()

        return storage

    def pack_session_data_to_string(self, data: SessionData) -> str:
        packed = struct.pack(
            pylogram.storage.Storage.SESSION_STRING_FORMAT,
            data.dc_id,
            0,
            data.test_mode,
            data.auth_key,
            data.user_id,
            data.is_bot
        )

        return base64.urlsafe_b64encode(packed).decode().rstrip("=")



from typing import Optional


def parse_none_if_null_str(value: str) -> Optional[str]:
    if value is None:
        return None

    if isinstance(value, str):
        if value == 'None' or value == 'null' or value:
            return None

    return value



import random

import pydantic

from ks_session_manager.env import env
from ks_session_manager.types import SessionProxy
from .phone import guess_country_by_phone_number


class SmartProxy(pydantic.BaseModel):
    country: str = "random"
    hostname: str = 'gate.smartproxy.com'
    sticky_port_min: int = 10000
    sticky_port_max: int = 10019
    rotating_port: int = 10000
    username: str | None = env.SMART_PROXY_USERNAME
    password: str | None = env.SMART_PROXY_PASSWORD


SMART_PROXY_PROXIES: dict[str, SmartProxy] = {
    "russia": SmartProxy(
        hostname="ru.smartproxy.com",
        sticky_port_min=40001,
        sticky_port_max=40010,
        rotating_port=40000
    ),
    "thailand": SmartProxy(
        hostname="th.smartproxy.com",
        sticky_port_min=30001,
        sticky_port_max=30010,
        rotating_port=30000
    ),
    "iran": SmartProxy(
        hostname="in.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000
    ),
    "pakistan": SmartProxy(
        hostname="pk.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000
    ),
    "malaysia": SmartProxy(
        hostname="my.smartproxy.com",
        sticky_port_min=30001,
        sticky_port_max=30010,
        rotating_port=30000
    ),
    "united_kingdom": SmartProxy(
        hostname="gb.smartproxy.com",
        sticky_port_min=30001,
        sticky_port_max=30010,
        rotating_port=30000
    ),
    # "iran": SmartProxy(
    #     hostname="ir.smartproxy.com",
    #     sticky_port_min=30001,
    #     sticky_port_max=30010,
    #     rotating_port=30000
    # ),
    "indonesia": SmartProxy(
        hostname="id.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000
    ),
    "vietnam": SmartProxy(
        hostname="vn.smartproxy.com",
        sticky_port_min=46001,
        sticky_port_max=46010,
        rotating_port=46000
    ),
    "kazakhstan": SmartProxy(
        hostname="kz.smartproxy.com",
        sticky_port_min=40001,
        sticky_port_max=40010,
        rotating_port=40000
    ),
    "usa": SmartProxy(
        hostname="us.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000
    ),
    "canada": SmartProxy(
        hostname="ca.smartproxy.com",
        sticky_port_min=20001,
        sticky_port_max=20010,
        rotating_port=20000
    ),
    "india": SmartProxy(
        hostname="in.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000
    ),
    "kenya": SmartProxy(
        hostname="ke.smartproxy.com",
        sticky_port_min=45001,
        sticky_port_max=45010,
        rotating_port=45000
    ),
    "philippines": SmartProxy(
        hostname="ph.smartproxy.com",
        sticky_port_min=40001,
        sticky_port_max=40010,
        rotating_port=40000
    ),
    "south_africa": SmartProxy(
        hostname="za.smartproxy.com",
        sticky_port_min=40001,
        sticky_port_max=40010,
        rotating_port=40000
    ),
    "senegal": SmartProxy(
        hostname="sn.smartproxy.com",
        sticky_port_min=49001,
        sticky_port_max=490010,
        rotating_port=49000
    ),
    "romania": SmartProxy(
        hostname="ro.smartproxy.com",
        sticky_port_min=13001,
        sticky_port_max=13010,
        rotating_port=13000
    ),
    "morocco": SmartProxy(
        hostname="ma.smartproxy.com",
        sticky_port_min=40001,
        sticky_port_max=40010,
        rotating_port=40000
    ),
    "algeria": SmartProxy(
        hostname="gate.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000,
        username="user-kservice-country-dz"
    ),
    "tajikistan": SmartProxy(
        hostname="gate.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000,
        username="user-kservice-country-tj"
    ),
    "brazil": SmartProxy(
        hostname="br.smartproxy.com",
        sticky_port_min=10001,
        sticky_port_max=10010,
        rotating_port=10000,
    ),
    "mexico": SmartProxy(
        hostname="mx.smartproxy.com",
        sticky_port_min=20001,
        sticky_port_max=20010,
        rotating_port=20000,
    ),

}
DEFAULT_SMART_PROXY = SmartProxy(
    hostname="gate.smartproxy.com",
    sticky_port_min=10001,
    sticky_port_max=10010,
    rotating_port=7000
)


async def get_proxy_by_country(country: str, *, rotating: bool = False) -> SessionProxy:
    """Fetches a proxy by country.

    Args:
        country (str): The country of the desired proxy.
        rotating (bool, optional): Whether to use a rotating proxy. Defaults to False.

    Returns:
        SessionProxy: The fetched proxy.

    """
    smartproxy_by_country = SMART_PROXY_PROXIES.get(country)

    if not bool(smartproxy_by_country):
        smartproxy_by_country = DEFAULT_SMART_PROXY

    if rotating:
        port = smartproxy_by_country.rotating_port
    else:
        port = random.randint(smartproxy_by_country.sticky_port_min, smartproxy_by_country.sticky_port_max)

    return SessionProxy(
        scheme='http',
        hostname=smartproxy_by_country.hostname,
        port=port,
        username=smartproxy_by_country.username,
        password=smartproxy_by_country.password,
    )


async def get_proxy_by_phone_number(phone_number: str, *, rotating: bool = False) -> SessionProxy:
    """
    Get a proxy based on the given phone number.

    Parameters:
    - phone_number (str): The phone number used to determine the country.
    - rotating (bool, optional): Flag indicating whether to use a rotating proxy. Default is False.

    Returns:
    - SessionProxy: The proxy obtained based on the phone number and country.

    """
    phone_country = guess_country_by_phone_number(phone_number)
    return await get_proxy_by_country(phone_country, rotating=rotating)



from __future__ import annotations

import abc
import asyncio
import logging
import pathlib
from typing import Any

from ks_session_manager.types import SessionData
from ks_session_manager.types import SessionMetadata


async def _read_2fa(file_path: str | pathlib.Path) -> str | None:
    file_path = pathlib.Path(file_path)
    two_fa_file_names = [
        'pass', 'pass.txt',
        'password', 'password.txt',
        '2FA', '2FA.txt',
        '2fa', '2fa.txt',
        'Password2FA', 'Password2FA.txt',
        'twoFA', 'twoFA.txt',
        'twofa', 'twofa.txt',
    ]
    two_fa_file = next((
        file_path.with_name(file_name)
        for file_name in two_fa_file_names
        if file_path.with_name(file_name).exists()
    ), None)

    if two_fa_file:
        return await asyncio.to_thread(two_fa_file.read_text, encoding='utf-8', errors=None)

    return None


class BaseSessionFileManager(abc.ABC):
    def __init__(self, session_name: str, *, workdir: str | pathlib.Path | None = None):
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        self.session_name = session_name
        self.workdir = pathlib.Path(workdir).absolute() if workdir is not None else pathlib.Path.cwd().absolute()

    @property
    @abc.abstractmethod
    def session_file_extension(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    async def read_session_storage(self) -> SessionData:
        raise NotImplementedError

    @abc.abstractmethod
    async def write_session_storage(
            self,
            session_data: SessionData,
            *,
            in_memory: bool = False,
            exist_ok: bool = False
    ) -> Any:
        raise NotImplementedError

    @property
    def session_path(self) -> pathlib.Path:
        return (self.workdir / self.session_name).with_suffix(self.session_file_extension).absolute()

    @property
    def session_json_metadata_path(self) -> pathlib.Path:
        return self.session_path.with_suffix('.json').absolute()

    async def read_session_metadata(self) -> SessionMetadata:
        if not self.session_json_metadata_path.exists():
            raise FileNotFoundError(f"Session metadata file not found in {self.session_json_metadata_path}")

        if not self.session_json_metadata_path.is_file():
            raise IsADirectoryError(
                f"Session metadata file path "
                f"{self.session_json_metadata_path} points to a directory"
            )

        # noinspection PyTypeChecker
        metadata_json = await asyncio.to_thread(self.session_json_metadata_path.read_text, encoding='utf-8')
        session_metadata = SessionMetadata.model_validate_json(metadata_json)

        if not bool(session_metadata.two_factor_password):
            session_metadata.two_factor_password = await _read_2fa(self.session_json_metadata_path)

        return session_metadata

    async def write_session_metadata(self, session_metadata: SessionMetadata):
        session_metadata_file_path = self.session_json_metadata_path

        if session_metadata_file_path.exists():
            raise FileExistsError(f"Metadata JSON file is already exists in {session_metadata_file_path}")

        metadata_json = session_metadata.model_dump_json(indent=4, by_alias=True)
        # noinspection PyTypeChecker
        await asyncio.to_thread(session_metadata_file_path.write_text, metadata_json, encoding='utf-8')

    async def delete_session(self, *, i_accept_possible_lost_of_data: bool = False):
        if not i_accept_possible_lost_of_data:
            self.logger.warning(
                "You are about to delete session. Pass i_accept_possible_lost_of_data=True to remove session"
            )
            return

        self.session_path.unlink(missing_ok=True)
        self.session_json_metadata_path.unlink(missing_ok=True)



from __future__ import annotations

from typing import Any

import telethon.sessions
import telethon.types
from telethon import crypto as telethon_crypto

from ks_session_manager.managers.base_manager import BaseSessionFileManager
from ks_session_manager.types import SessionData
from ks_session_manager.utils.parsers import parse_none_if_null_str


class TelethonSessionFileManager(BaseSessionFileManager):
    @property
    def session_file_extension(self) -> str:
        return telethon.sessions.sqlite.EXTENSION

    def _get_file_storage(self, *, exists: bool = False) -> telethon.sessions.SQLiteSession:
        if not exists and self.session_path.exists():
            raise FileExistsError(f"Session storage is already exists in {self.session_path}")

        if exists and not self.session_path.exists():
            raise FileNotFoundError(f"Session storage not found in {self.session_path}")

        if self.session_path.is_dir():
            raise IsADirectoryError(f"Session file path {self.session_path} points to a directory")

        return telethon.sessions.SQLiteSession(str(self.session_path))

    async def read_session_storage(self) -> SessionData:
        session = self._get_file_storage(exists=True)
        # Telethon feature to save current user id in entities
        self_entity = session.get_entity_rows_by_id(0)

        return SessionData(
            dc_id=session.dc_id,
            user_id=self_entity[1] if bool(self_entity) else None,
            auth_key=session.auth_key.key,
            server_address=parse_none_if_null_str(session.server_address),
            server_port=session.port,
            takeout_id=parse_none_if_null_str(session.takeout_id),
        )

    async def write_session_storage(
            self,
            data: SessionData,
            *,
            in_memory: bool = False,
            exist_ok: bool = False
    ) -> Any:
        if not bool(data.user_id):
            raise ValueError("user_id must be set in session data")

        if in_memory:
            storage = telethon.sessions.MemorySession()
        else:
            try:
                storage = self._get_file_storage(exists=False)
            except FileExistsError:
                if exist_ok:
                    return self._get_file_storage(exists=True)

                raise

        storage.set_dc(data.dc_id, data.server_address, data.server_port)
        storage.auth_key = telethon_crypto.AuthKey(data.auth_key)
        storage.takeout_id = data.takeout_id
        # noinspection PyTypeChecker
        current_user_entity = telethon.types.contacts.ResolvedPeer(
            None,
            [telethon.types.InputPeerUser(0, data.user_id)],
            []
        )
        storage.process_entities(current_user_entity)
        storage.save()
        storage.close()
        return storage



class ConverterError(Exception):
    pass


class SigningInDisabledError(RuntimeError):
    def __init__(self, code_type: str):
        super().__init__("Session is not signed in")
        self.code_type = code_type


class ConverterSessionUnauthorized(ConverterError):
    def __init__(self, message: str, original_exception: Exception):
        super().__init__(message)
        self.original_exception = original_exception


class ConverterSessionUnauthorizedYetError(ConverterError):
    pass


class ConverterMissingSessionData(ConverterError):
    pass


class ConverterUnableToCollectUserInfo(ConverterError):
    pass



import abc
import typing

import pylogram.errors
import telethon.errors

from ks_session_manager.errors import ConverterSessionUnauthorizedYetError
from ks_session_manager.types import SessionUserInfo
from .pylogram import PylogramClient
from .telethon import TelethonClient
from ..errors import ConverterSessionUnauthorized


class BaseTelegramAdapter(abc.ABC):
    @property
    @abc.abstractmethod
    def is_connected(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_user_info(self) -> SessionUserInfo:
        raise NotImplementedError

    @abc.abstractmethod
    async def start(self):
        raise NotImplementedError

    @abc.abstractmethod
    async def stop(self):
        raise NotImplementedError

    async def __aenter__(self) -> typing.Self:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            await self.stop()
        except Exception:
            pass

        if exc_val:
            raise exc_val


class TelethonTelegramAdapter(BaseTelegramAdapter):
    def __init__(self, client: TelethonClient):
        self.client = client

    @property
    def is_connected(self) -> bool:
        return self.client.is_connected()

    async def get_user_info(self) -> SessionUserInfo:
        if not self.is_connected:
            raise RuntimeError('Client is not connected')

        me = await self.client.get_me()

        if me is None:
            raise ConverterSessionUnauthorizedYetError

        return SessionUserInfo(
            id=self.client.me.id,
            first_name=self.client.me.first_name,
            last_name=self.client.me.last_name,
            username=self.client.me.username,
            phone_number=self.client.me.phone,
            is_premium=self.client.me.premium,
        )

    async def start(self):
        # noinspection PyUnresolvedReferences
        await self.client.connect()

    async def stop(self):
        await self.client.disconnect()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            await self.stop()
        except Exception:
            pass

        if isinstance(exc_val, (telethon.errors.UnauthorizedError, telethon.errors.AuthKeyError)):
            raise ConverterSessionUnauthorized(str(exc_val), original_exception=exc_val) from exc_val


class PylogramTelegramAdapter(BaseTelegramAdapter):
    def __init__(self, client: PylogramClient):
        self.client = client

    @property
    def is_connected(self) -> bool:
        return self.client.is_connected

    async def get_user_info(self) -> SessionUserInfo:
        if not self.is_connected:
            raise RuntimeError('Client is not connected')

        return SessionUserInfo(
            id=self.client.me.id,
            first_name=self.client.me.first_name,
            last_name=self.client.me.last_name,
            username=self.client.me.username,
            phone_number=self.client.me.phone_number,
            is_premium=self.client.me.is_premium,
        )

    async def start(self):
        try:
            await self.client.start(authorize=False)
        except RuntimeError as e:
            raise ConverterSessionUnauthorizedYetError from e

    async def stop(self):
        await self.client.stop()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            await self.stop()
        except Exception:
            pass

        if isinstance(exc_val, (pylogram.errors.Unauthorized, pylogram.errors.AuthKeyDuplicated)):
            raise ConverterSessionUnauthorized(str(exc_val), original_exception=exc_val) from exc_val



import asyncio
import logging
import random
import re
from datetime import datetime
from datetime import timezone
from pathlib import Path

import telethon
import telethon.sessions

from ks_session_manager.types import SessionMetadata
from ks_session_manager.types import SessionProxy

logger = logging.getLogger(__name__)


class TelethonClient(telethon.TelegramClient):
    def __init__(
            self,
            data: SessionMetadata,
            *,
            session_path: str | Path | None = None,
            session_storage: telethon.sessions.Session | None = None,
            proxy: SessionProxy | None = None,
            sign_in_enabled: bool = True,
            # Passthrough
            connection: type[telethon.network.Connection] = telethon.network.ConnectionTcpFull,
            use_ipv6: bool = False,
            local_addr: str | tuple = None,
            timeout: int = 10,
            request_retries: int = 5,
            connection_retries: int = 5,
            retry_delay: int = 1,
            auto_reconnect: bool = True,
            sequential_updates: bool = False,
            flood_sleep_threshold: int = 60,
            raise_last_call_error: bool = False,
            loop: asyncio.AbstractEventLoop = None,
            base_logger: str | logging.Logger = None,
            receive_updates: bool = True,
            entity_cache_limit: int = 5000,
            catch_up: bool = False,
    ):
        if not bool(session_path) and not bool(session_storage):
            raise ValueError("Either session_path or session_storage must be provided")

        if bool(session_path) and bool(session_storage):
            raise ValueError("Either session_path or session_storage must be provided, not both")

        if bool(session_path):
            session = str(Path(session_path).resolve())
        else:
            session = session_storage

        self.session: telethon.sessions.Session | None = None
        self.me: telethon.types.User | None = None
        self.session_metadata: SessionMetadata = data
        self.sign_in_enabled = sign_in_enabled

        super().__init__(
            session,
            data.api_id,
            data.api_hash,
            proxy=proxy.to_telethon_format() if bool(proxy) else None,
            device_model=data.device_model,
            system_version=data.system_version,
            app_version=data.app_version,
            lang_code=data.lang_code,
            system_lang_code=data.system_lang_code,
            # Other
            catch_up=catch_up,
            # Passthrough
            connection=connection,
            use_ipv6=use_ipv6,
            local_addr=local_addr,
            timeout=timeout,
            request_retries=request_retries,
            connection_retries=connection_retries,
            retry_delay=retry_delay,
            auto_reconnect=auto_reconnect,
            sequential_updates=sequential_updates,
            flood_sleep_threshold=flood_sleep_threshold,
            raise_last_call_error=raise_last_call_error,
            loop=loop,
            base_logger=base_logger,
            receive_updates=receive_updates,
            entity_cache_limit=entity_cache_limit
        )
        # noinspection PyUnresolvedReferences
        self._init_request.lang_pack = data.lang_pack

    async def _on_login(self, user: telethon.types.User | None):
        self.me = await super()._on_login(user)
        logger.info(f"Signed in as {self.me.id}: {self.me.phone}")

    async def export_session_string(self) -> str:
        storage = telethon.sessions.StringSession()
        storage.set_dc(
            self.session.dc_id,
            self.session.server_address,
            self.session.port
        )
        storage.auth_key = self.session.auth_key
        return storage.save()

    async def retrieve_telegram_code(self, *, request_date: datetime = None, timeout: int = 60) -> str | None:
        """
        Retrieves the verification code from a Telegram message.

        Parameters:
        - request_date: Optional datetime parameter representing the date and time from which the code should be retrieved. Defaults to the current date and time in UTC if not provided.
        - timeout: Optional integer parameter representing the maximum time in seconds to wait for a code to be retrieved. Defaults to 60 seconds if not provided.

        Returns:
        - A string representing the retrieved verification code if successful.
        - None if no code is found within the specified timeout period.

        """
        if request_date is None:
            request_date = datetime.now(tz=timezone.utc)

        async with asyncio.timeout(timeout):
            while True:
                async for message in self.iter_messages(777000, limit=1):
                    if int(message.date.timestamp()) < int(request_date.timestamp()):
                        break

                    try:
                        return re.findall(r'[0-9]{5}', message.text, flags=re.MULTILINE)[0]
                    except IndexError:
                        logging.warning("Found new Telegram message but no code received")

                await asyncio.sleep(random.uniform(1, 3))



from __future__ import annotations

import logging
import pathlib
from datetime import datetime
from datetime import timezone

import pylogram.session

from ks_session_manager.clients.adapters import PylogramTelegramAdapter
from ks_session_manager.clients.adapters import TelethonTelegramAdapter
from ks_session_manager.clients.pylogram import PylogramClient
from ks_session_manager.clients.telethon import TelethonClient
from ks_session_manager.env import env
from ks_session_manager.errors import ConverterMissingSessionData
from ks_session_manager.errors import ConverterSessionUnauthorizedYetError
from ks_session_manager.errors import ConverterUnableToCollectUserInfo
from ks_session_manager.managers import BaseSessionFileManager
from ks_session_manager.managers import create_session_manager_of_type
from ks_session_manager.types import AccountType
from ks_session_manager.types import SessionData
from ks_session_manager.types import SessionMetadata
from ks_session_manager.types import SessionProxy
from ks_session_manager.types import TelegramPlatform
from ks_session_manager.utils.metadata_generator import MetadataGenerator


class Converter:
    def __init__(
            self,
            *,
            session_metadata: SessionMetadata | None = None,
            session_data: SessionData | None = None,
            proxy: SessionProxy | None = None
    ):
        self.session_data = session_data
        self.session_metadata = session_metadata
        self.proxy = proxy
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

    @classmethod
    async def create(
            cls,
            account_type: AccountType,
            session_name: str,
            *,
            workdir: str | pathlib.Path | None = None,
            proxy: SessionProxy | None = None,
            generate_metadata_if_not_exists: bool = False,
            generate_metadata_platform: TelegramPlatform = TelegramPlatform.DESKTOP_LINUX,
            allow_online: bool = False
    ) -> Converter:
        """
        Create a converter object for the specified account type and session name.

        Parameters:
            account_type (AccountType): The type of the account.
            session_name (str): The name of the session.
            workdir (str | pathlib.Path | None, optional): The working directory for the session. Defaults to None.
            proxy (SessionProxy | None, optional): The session proxy. Defaults to None.
            generate_metadata_if_not_exists (bool, optional): Whether to generate session metadata if it doesn't exist. Defaults to False.
            generate_metadata_platform (TelegramPlatform, optional): The platform for generating metadata. Defaults to TelegramPlatform.DESKTOP_LINUX.
            allow_online (bool, optional): Whether to allow online mode to collect missing session data. Defaults to False.

        Returns:
            Converter: The created converter object.

        Raises:
            NotImplementedError: If online mode is enabled.

        """
        manager: BaseSessionFileManager = create_session_manager_of_type(account_type, session_name, workdir=workdir)
        session_data = await manager.read_session_storage()

        try:
            session_metadata = await manager.read_session_metadata()
        except FileNotFoundError:
            if allow_online:
                raise NotImplementedError("Online mode is not implemented yet")
            elif generate_metadata_if_not_exists:
                session_metadata = MetadataGenerator(generate_metadata_platform, session_name).generate()
            else:
                raise

        if not bool(session_data.user_id) and bool(session_metadata.user_id):
            session_data.user_id = session_metadata.user_id

        if not bool(session_data.user_id):
            # Ensure user_id is set, load online if allowed
            if not allow_online:
                raise ConverterMissingSessionData(
                    "Session data is incomplete and online mode is not allowed. User ID is not set. "
                    "Enable online mode to collect missing data."
                )

            try:
                if account_type == AccountType.TELETHON:
                    adapter = TelethonTelegramAdapter(
                        client=TelethonClient(
                            data=session_metadata,
                            session_path=manager.session_path,
                            proxy=proxy,
                            sign_in_enabled=False
                        )
                    )
                elif account_type == AccountType.PYLOGRAM:
                    adapter = PylogramTelegramAdapter(
                        client=PylogramClient(
                            data=session_metadata,
                            session_path=manager.session_path,
                            proxy=proxy,
                            sign_in_enabled=False
                        )
                    )
                else:
                    raise ValueError(f"Unsupported account type: {account_type}")
            except Exception as e:
                raise ConverterUnableToCollectUserInfo(
                    f"Session data is incomplete and unable to collect missing data: {e}"
                ) from e

            try:
                async with adapter:
                    user_info = await adapter.get_user_info()
            except ConverterSessionUnauthorizedYetError as e:
                raise e
            except Exception as e:
                raise ConverterUnableToCollectUserInfo(
                    "Session data is incomplete and unable to collect missing data. "
                ) from e

            session_data.user_id = user_info.id
            session_metadata.user_id = user_info.id
            session_metadata.phone_number = user_info.phone_number
            session_metadata.first_name = user_info.first_name
            session_metadata.last_name = user_info.last_name
            session_metadata.username = user_info.username
            session_metadata.is_premium = user_info.is_premium
            session_metadata.last_check_time = int(datetime.now(tz=timezone.utc).timestamp())

        return cls(session_metadata=session_metadata, session_data=session_data, proxy=proxy)

    @classmethod
    async def from_telethon(
            cls,
            session_name: str,
            *,
            workdir: str | pathlib.Path | None = env.TELETHON_WORKDIR,
            proxy: SessionProxy | None = None,
            generate_metadata_if_not_exists: bool = False,
            generate_metadata_platform: TelegramPlatform = TelegramPlatform.DESKTOP_LINUX,
            allow_online: bool = False
    ) -> Converter:
        """

        Converts a Telethon session into a Converter instance.

        Parameters:
            session_name (str): The name of the Telethon session to convert.
            workdir (str | pathlib.Path | None, optional): The directory where the converted session file will be stored. Defaults to env.TELETHON_WORKDIR.
            proxy (SessionProxy | None, optional): The proxy to be used for the converted session. Defaults to None.
            generate_metadata_if_not_exists (bool, optional): Whether to generate metadata for the converted session if it doesn't exist. Defaults to False.
            generate_metadata_platform (TelegramPlatform, optional): The platform to be used for generating metadata for the converted session. Defaults to TelegramPlatform.DESKTOP_LINUX.
            allow_online (bool, optional): Whether the converted session should be used in online mode. Defaults to False.

        Returns:
            Converter: The Converter instance created from the Telethon session.

        """
        return await cls.create(
            account_type=AccountType.TELETHON,
            session_name=session_name,
            workdir=workdir,
            proxy=proxy,
            generate_metadata_if_not_exists=generate_metadata_if_not_exists,
            generate_metadata_platform=generate_metadata_platform,
            allow_online=allow_online
        )

    @classmethod
    async def from_pylogram(
            cls,
            session_name: str,
            *,
            workdir: str | pathlib.Path | None = env.PYLOGRAM_WORKDIR,
            proxy: SessionProxy | None = None,
            generate_metadata_if_not_exists: bool = False,
            generate_metadata_platform: TelegramPlatform = TelegramPlatform.DESKTOP_LINUX,
            allow_online: bool = False
    ) -> Converter:
        """

        Converts a Pyrogram session into a Converter instance.

        Parameters:
            session_name (str): The name of the Pyrogram session.
            workdir (str | pathlib.Path | None): The directory path where session files will be stored. Defaults to env.PYLOGRAM_WORKDIR.
            proxy (SessionProxy | None): The session proxy to be used. Defaults to None.
            generate_metadata_if_not_exists (bool): Whether to generate session metadata if it doesn't exist. Defaults to False.
            generate_metadata_platform (TelegramPlatform): The platform used to generate session metadata. Defaults to TelegramPlatform.DESKTOP_LINUX.
            allow_online (bool): Whether to create an online session. Defaults to False.

        Returns:
            Converter: An instance of the Converter class.

        """
        return await cls.create(
            account_type=AccountType.PYLOGRAM,
            session_name=session_name,
            workdir=workdir,
            proxy=proxy,
            generate_metadata_if_not_exists=generate_metadata_if_not_exists,
            generate_metadata_platform=generate_metadata_platform,
            allow_online=allow_online
        )

    async def export_session(
            self,
            account_type: AccountType,
            session_name: str | None = None,
            workdir: str | pathlib.Path | None = None,
            export_metadata: bool = True
    ):
        """

        Export a session to the specified account type, session name, and working directory.

        :param account_type: The account type to export the session to.
        :type account_type: AccountType
        :param session_name: The name of the session to be exported. If not provided, it defaults to the phone number associated with the current session.
        :type session_name: str, optional
        :param workdir: The working directory where the session should be exported. If not provided, it defaults to the current working directory.
        :type workdir: str or pathlib.Path, optional
        :param export_metadata: Indicates whether to export the session metadata alongside the session data. Defaults to True.
        :type export_metadata: bool, optional
        """
        if not bool(session_name):
            session_name = self.session_metadata.phone_number

        manager = create_session_manager_of_type(account_type, session_name, workdir=workdir)
        await manager.write_session_storage(self.session_data, in_memory=False, exist_ok=False)
        self.logger.info(f"Session successfully exported to {manager.session_path}")

        if export_metadata:
            await manager.write_session_metadata(self.session_metadata)
            self.logger.info(f"Session metadata successfully exported to {manager.session_json_metadata_path}")

    async def export_telethon_session(
            self,
            session_name: str = None,
            workdir: str | pathlib.Path = env.TELETHON_WORKDIR,
            export_metadata: bool = True
    ):
        """
        Exports the Telethon session to a specified file or directory.

        :param session_name: Optional name of the session file. If not provided, a default name will be used.
        :type session_name: str, optional
        :param workdir: Optional working directory where the session file will be exported. If not provided, the default work directory will be used.
        :type workdir: str or pathlib.Path, optional
        :param export_metadata: Optional flag indicating whether to export the session metadata. Default is True.
        :type export_metadata: bool, optional
        :return: None
        :rtype: None
        """
        await self.export_session(AccountType.TELETHON, session_name, workdir, export_metadata)

    async def export_pylogram_session(
            self,
            session_name: str = None,
            workdir: str | pathlib.Path = env.PYLOGRAM_WORKDIR,
            export_metadata: bool = True
    ):
        """
        Export a Pylogram session to a specified file or directory.

        :param session_name: Optional name of the session file. If not provided, a default name will be used.
        :type session_name: str, optional
        :param workdir: Optional working directory where the session file will be exported. If not provided, the default work directory will be used.
        :type workdir: str or pathlib.Path, optional
        :param export_metadata: Optional flag indicating whether to export the session metadata. Default is True.
        :type export_metadata: bool, optional

        :return: None
        :rtype: None

        """
        await self.export_session(AccountType.PYLOGRAM, session_name, workdir, export_metadata)

    async def get_telethon_client(
            self,
            *,
            in_memory: bool = False,
            proxy: SessionProxy | None = None,
            workdir: str | pathlib.Path = env.TELETHON_WORKDIR,
            **telethon_client_kwargs
    ) -> TelethonClient:
        """
        Asynchronously retrieves a TelethonClient instance with the specified parameters.

        Parameters:
        - in_memory (bool): Whether to store the session storage in memory. Defaults to False.
        - proxy (SessionProxy | None): The session proxy to use. Defaults to None.
        - workdir (str | pathlib.Path): The directory path for storing session data. Defaults to env.TELETHON_WORKDIR.

        Returns:
        - TelethonClient: The instantiated TelethonClient object.

        Example Usage:
        ```python
        client = await get_telethon_client(in_memory=True, proxy=my_proxy, workdir="/path/to/directory")
        ```
        """
        session_name = self.session_metadata.phone_number
        manager = create_session_manager_of_type(AccountType.TELETHON, session_name, workdir=workdir)
        session_storage = await manager.write_session_storage(self.session_data, in_memory=in_memory, exist_ok=True)

        if not bool(proxy):
            proxy = self.proxy

        return TelethonClient(
            self.session_metadata,
            session_storage=session_storage,
            proxy=proxy,
            **telethon_client_kwargs
        )

    # Pylogram
    async def get_pylogram_client(
            self,
            *,
            in_memory: bool = False,
            proxy: SessionProxy | None = None,
            workdir: str | pathlib.Path = env.PYLOGRAM_WORKDIR,
            **pylogram_client_kwargs
    ) -> PylogramClient:
        """

        Asynchronously creates and returns a PylogramClient object.

        Parameters:
        - `in_memory` : bool, optional (default=False)
            If True, the session storage will be stored in memory instead of on disk.

        - `proxy` : SessionProxy or None, optional (default=None)
            The proxy to be used for the PylogramClient object. If None, the proxy from the session metadata will be used.

        - `workdir` : str or pathlib.Path, optional (default=env.PYLOGRAM_WORKDIR)
            The directory path where the session storage will be stored.

        Returns:
        - `PylogramClient` : The PylogramClient object that is created.

        """
        session_name = self.session_metadata.phone_number
        manager = create_session_manager_of_type(AccountType.PYLOGRAM, session_name, workdir=workdir)
        session_storage = await manager.write_session_storage(self.session_data, in_memory=in_memory, exist_ok=True)

        if not bool(proxy):
            proxy = self.proxy

        return PylogramClient(
            self.session_metadata,
            session_storage=session_storage,
            proxy=proxy,
            **pylogram_client_kwargs
        )

    async def get_pylogram_session_string(self) -> str:
        client = await self.get_pylogram_client(in_memory=True)
        storage: pylogram.storage.Storage = client.storage
        await storage.open()
        session_string = await storage.export_session_string()
        await storage.close()
        return session_string

    async def get_telethon_session_string(self) -> str:
        client = await self.get_telethon_client(in_memory=True)
        return await client.export_session_string()

    @property
    def self_user_id(self) -> int | None:
        return self.session_data.user_id or self.session_metadata.user_id



import asyncio
import logging
import pathlib
from pathlib import Path

import pydantic
import pydash
import telethon
import telethon.client
import telethon.helpers
import telethon.sessions

from ks_session_manager.clients.telethon import TelethonClient
from ks_session_manager.errors import ConverterSessionUnauthorizedYetError
from ks_session_manager.types import SessionMetadata
from ks_session_manager.types import SessionProxy
from ks_session_manager.utils import smartproxy

logger = logging.getLogger(__name__)


class TelethonRevizor:
    def __init__(self, threads: int = 1, *, dry_run: bool = False):
        self.threads = threads
        self.semaphore = asyncio.BoundedSemaphore(threads)
        self.logger = logging.getLogger(__name__)
        self.dry_run = dry_run

    async def revise_session_in_path(self, session_path: str | Path, proxy: SessionProxy | None = None):
        """
        Revise the session located at the given path.

        Parameters:
        - session_path: A string or Path object representing the path to the session file.
        - proxy: A SessionProxy object representing the proxy to be used. (default: None)

        Raises:
        - FileNotFoundError: If the provided path is not a file.
        - ValueError: If the provided path does not have a .session extension.
        - RuntimeError: If the metadata file is invalid.

        Returns: None
        """
        session_path = Path(session_path).resolve()

        async with self.semaphore:
            metadata_file = session_path.with_suffix(".json")

            if not session_path.is_file() or not metadata_file.is_file():
                raise FileNotFoundError("Provided path must be a file")

            if session_path.suffix != ".session":
                raise ValueError("Provided path must be a .session file")

            try:
                metadata = SessionMetadata.model_validate_json(metadata_file.read_text())
            except pydantic.ValidationError as e:
                raise RuntimeError(f"Invalid metadata file: {e}")

            if not bool(proxy):
                proxy = await smartproxy.get_proxy_by_phone_number(metadata.phone_number, rotating=True)
                self.logger.info("Using proxy: %s", proxy.link)

            self.logger.info(f"Session revision started for session {session_path.stem}")
            telethon_client = TelethonClient(session_path, metadata, proxy=proxy)
            dir_to_move: pathlib.Path | None = None

            if not self.dry_run:
                try:
                    # noinspection PyUnresolvedReferences
                    await telethon_client.start(
                        phone=lambda: session_path.stem,
                        password=lambda: metadata.two_factor_password,
                    )
                except telethon.errors.PhoneNumberInvalidError:
                    self.logger.error(f"Session {session_path.stem} phone number invalid")
                    dir_to_move = session_path.parent / "phone_number_invalid"
                except telethon.errors.PhoneNumberBannedError:
                    self.logger.error(f"Session {session_path.stem} phone number banned")
                    dir_to_move = session_path.parent / "phone_number_banned"
                except telethon.errors.UnauthorizedError as e:
                    self.logger.error(f"Session {session_path.stem} unauthorized: {e}")
                    dir_to_move = session_path.parent / pydash.snake_case(e.__class__.__name__.replace('_error', ''))
                except telethon.errors.AuthKeyError as e:
                    self.logger.error(f"Session {session_path.stem} auth key error")
                    dir_to_move = session_path.parent / pydash.snake_case(e.__class__.__name__.replace('_error', ''))
                except ConverterSessionUnauthorizedYetError:
                    self.logger.error(f"Session {session_path.stem} is not signed in")
                    dir_to_move = session_path.parent / "not_signed_in"
                except ConnectionError:
                    self.logger.error(f"Session {session_path.stem} connection error")
                    dir_to_move = session_path.parent / "connection_error"
                except Exception as e:
                    self.logger.error(f"Session {session_path.stem} unhandled error: {e.__class__.__name__}: {e}")
                    dir_to_move = None
                else:
                    self.logger.info(f"Session {session_path.stem} is alive")
                    self.logger.info(f"Authorized as {telethon_client.me.id}: {telethon_client.me.phone}")
                    dir_to_move = session_path.parent / "alive"
                finally:
                    await telethon_client.disconnect()

            if dir_to_move:
                dir_to_move.mkdir(parents=True, exist_ok=True)
                session_path.rename(dir_to_move / session_path.name)
                metadata_file.rename(dir_to_move / metadata_file.name)

            self.logger.info(f"Session revision finished for session {session_path.stem}")

    async def revise_all_in_path(self, path: str | Path, *, limit: int | None = None):
        """
        Revise All in Path

        Revise all session files in a given directory path.

        Parameters:
        - path: str | Path - The directory path containing the session files to be revised.
        - limit: int | None - Optional. The maximum number of session files to be revised. Default is None, which means to revise all session files in the directory.

        Raises:
        - NotADirectoryError: If the provided path is not a directory.

        Example usage:
        ```python
        await revise_all_in_path('/path/to/directory', limit=5)
        ```
        """
        path = Path(path).resolve()

        if not path.is_dir():
            raise NotADirectoryError("Provided path must be a directory")

        session_files = [
            file
            for file in path.iterdir()
            if file.is_file() and file.suffix == ".session"
        ]

        if limit:
            session_files = session_files[:limit]

        tasks = [self.revise_session_in_path(file) for file in session_files]

        self.logger.info(
            f"Starting {len(tasks)} revisions. "
            f"Simultaneous count: {self.threads}"
        )
        await asyncio.gather(*tasks)
        self.logger.info("All revisions finished")



from __future__ import annotations

import pathlib

from ks_session_manager.env import env
from ks_session_manager.types import AccountType
from .base_manager import BaseSessionFileManager
from .pylogram_manager import PylogramSessionFileManager
from .telethon_manager import TelethonSessionFileManager


def create_session_manager_of_type(
        account_type: AccountType,
        session_name: str,
        workdir: str | pathlib.Path | None = None
) -> BaseSessionFileManager:
    if account_type == AccountType.TELETHON:
        return TelethonSessionFileManager(session_name=session_name, workdir=workdir or env.TELETHON_WORKDIR)
    elif account_type == AccountType.PYLOGRAM:
        return PylogramSessionFileManager(session_name=session_name, workdir=workdir or env.PYLOGRAM_WORKDIR)
    else:
        raise TypeError(f"Account type {account_type} is not supported")



import logging
import tarfile
import zipfile

from pathlib import Path
from typing import Sequence

logger = logging.getLogger(__name__)


def recursive_remove_empty_dirs(directory: str | Path):
    """
    Removes all empty directories recursively within the specified directory.

    :param directory: The directory path to start removing empty directories from.
    :type directory: str or Path
    """
    directory = Path(directory)

    for file in directory.iterdir():
        if file.is_dir():
            recursive_remove_empty_dirs(file)
            if not any(file.iterdir()):
                file.rmdir()


def tar_filter(tarinfo: tarfile.TarInfo, _) -> tarfile.TarInfo | None:
    path = Path(tarinfo.name)

    # Skip hidden files
    if path.stem.startswith('.'):
        return None

    return tarinfo


def extract_archive(archive: str | Path, output_dir: str | Path):
    """
    Extracts the contents of an archive to a specified output directory.

    Parameters:
        archive (str | Path): Path to the archive file.
        output_dir (str | Path): Path to the output directory.

    Raises:
        ValueError: If the archive type is unsupported.

    """
    archive = Path(archive)
    output_dir = Path(output_dir)

    if tarfile.is_tarfile(archive):
        with tarfile.open(archive) as tar:
            tar.extractall(output_dir, filter=tar_filter)
    elif zipfile.is_zipfile(archive):
        with zipfile.ZipFile(archive) as file:
            file.extractall(output_dir)
    else:
        raise ValueError(f'{archive} has unsupported archive type')


def flatten_dir_files(
        directory: str | Path,
        *,
        include_suffixes: Sequence[str] | None = None,
        remove_unmatched: bool = False
):
    """
    Flatten Dir Files

    Flattens files in a given directory and its subdirectories, optionally matching only certain file suffixes and removing unmatched files.

    Parameters
    ----------
    directory : str or Path
        The directory to flatten files in.

    include_suffixes : Sequence[str]
        A sequence of file suffixes to include. Only files with these suffixes will be flattened.
        If an empty sequence is provided, all files will be included.

    remove_unmatched : bool, optional
        If True, unmatched files (files whose suffixes do not match the provided include_suffixes) will be removed.
        Default is False.

    Returns
    -------
    None

    Notes
    -----
    - Existing files with the same name as the flattened files will be skipped.
    - Empty directories will be cleaned after file flattening.

    Examples
    --------
    >>> flatten_dir_files('path/to/directory', include_suffixes=['.txt', '.csv'], remove_unmatched=True)
    """
    target_dir = Path(directory)

    for file in target_dir.glob('**/*'):
        if file.is_file():
            if bool(include_suffixes) and file.suffix not in include_suffixes:
                if remove_unmatched:
                    file.unlink()
                continue

            target_path = Path(target_dir / file.name)

            if target_path.exists():
                logger.warning(f'File {target_path} already exists, skipping')
                continue

            file.rename(target_path)

    # Clean empty directories
    logger.info(f"Cleaning empty directories in {target_dir}")
    recursive_remove_empty_dirs(target_dir)



from setuptools import find_packages
from setuptools import setup

with open('./ks_session_manager/__init__.py') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip("'")
            break


def read_requirements():
    with open('./requirements.txt') as f:
        req = f.read().splitlines()

    return req


with open('README.md', "r") as f:
    long_description = f.read()

setup(
    name='ks_session_manager',
    author="pylakey",
    version=version,
    packages=find_packages(),
    # license='MIT',
    # license_file='LICENSE',
    description='KS session manager',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=read_requirements(),
    python_requires='>=3.10',
)
